@@ -15,12 +15,13 @@ with initialize_s3_bucket.py.
 import contextlib
 import gzip
 import io
+import os
 import unittest
 import uuid
 import warnings
 
 import boto3
-from parameterizedtestcase import ParameterizedTestCase as PTestCase
+import pytest
 
 import smart_open
 import smart_open.concurrency
@@ -170,7 +171,7 @@ class ReaderTest(unittest.TestCase):
         expected = CONTENTS[key_name]
 
         with smart_open.s3.Reader(BUCKET_NAME, key_name) as fin:
-            returned_obj = fin.to_boto3()
+            returned_obj = fin.to_boto3(boto3.resource('s3'))
 
         boto3_body = returned_obj.get()['Body'].read()
         self.assertEqual(expected, boto3_body)
@@ -203,20 +204,53 @@ class WriterTest(unittest.TestCase):
 
     def test_multipart(self):
         """Does s3 multipart chunking work correctly?"""
-        with smart_open.s3.MultipartWriter(BUCKET_NAME, self.key, min_part_size=10) as fout:
-            fout.write(b"test")
-            self.assertEqual(fout._buf.tell(), 4)
+        data_dir = os.path.join(os.path.dirname(__file__), "../smart_open/tests/test_data")
+        with open(os.path.join(data_dir, "crime-and-punishment.txt"), "rb") as fin:
+            crime = fin.read()
+        data = b''
+        ps = 5 * 1024 * 1024
+        while len(data) < ps:
+            data += crime
 
-            fout.write(b"test\n")
-            self.assertEqual(fout._buf.tell(), 9)
-            self.assertEqual(fout._total_parts, 0)
+        title = "Преступление и наказание\n\n".encode()
+        to_be_continued = "\n\n... продолжение следует ...\n\n".encode()
 
-            fout.write(b"test")
-            self.assertEqual(fout._buf.tell(), 0)
-            self.assertEqual(fout._total_parts, 1)
+        key = "WriterTest.test_multipart"
+        with smart_open.s3.MultipartWriter(BUCKET_NAME, key, part_size=ps) as fout:
+            #
+            # Write some data without triggering an upload
+            #
+            fout.write(title)
+            assert fout._total_parts == 0
+            assert fout._buf.tell() == 48
 
-        data = read_key(self.key)
-        self.assertEqual(data, b"testtest\ntest")
+            #
+            # Trigger a part upload
+            #
+            fout.write(data)
+            assert fout._total_parts == 1
+            assert fout._buf.tell() == 661
+
+            #
+            # Write _without_ triggering a part upload
+            #
+            fout.write(to_be_continued)
+            assert fout._total_parts == 1
+            assert fout._buf.tell() == 710
+
+        #
+        # We closed the writer, so the final part must have been uploaded
+        #
+        assert fout._buf.tell() == 0
+        assert fout._total_parts == 2
+
+        #
+        # read back the same key and check its content
+        #
+        with smart_open.s3.open(BUCKET_NAME, key, 'rb') as fin:
+            got = fin.read()
+        want = title + data + to_be_continued
+        assert want == got
 
     def test_empty_key(self):
         """Does writing no data cause key with an empty value to be created?"""
@@ -277,7 +311,7 @@ def force(multiprocessing=False, concurrent_futures=False):
     smart_open.concurrency._CONCURRENT_FUTURES = old_concurrent_futures
 
 
-class IterBucketTest(PTestCase):
+class IterBucketTest(unittest.TestCase):
     def setUp(self):
         self.expected = [
             (key, value)
@@ -321,11 +355,17 @@ class IterBucketTest(PTestCase):
         self.assertEqual(len(expected), len(actual))
         self.assertEqual(expected, sorted(actual))
 
-    @PTestCase.parameterize(('workers',), [(x,) for x in (1, 4, 8, 16, 64)])
-    def test_workers(self, workers):
-        actual = list(smart_open.s3.iter_bucket(BUCKET_NAME, prefix='iter_bucket', workers=workers))
-        self.assertEqual(len(self.expected), len(actual))
-        self.assertEqual(self.expected, sorted(actual))
+
+@pytest.mark.parametrize('workers', [1, 4, 8, 16, 64])
+def test_workers(workers):
+    expected = sorted([
+        (key, value)
+        for (key, value) in CONTENTS.items()
+        if key.startswith('iter_bucket/')
+    ])
+    actual = sorted(smart_open.s3.iter_bucket(BUCKET_NAME, prefix='iter_bucket', workers=workers))
+    assert len(expected) == len(actual)
+    assert expected == actual
 
 
 class DownloadKeyTest(unittest.TestCase):
