@@ -51,11 +51,14 @@ class FakeBlobClient(object):
         self.__contents = io.BytesIO()
         self._staged_contents = {}
 
-    def commit_block_list(self, block_list):
+    def commit_block_list(self, block_list, metadata=None):
         data = b''.join([self._staged_contents[block_blob['id']] for block_blob in block_list])
         self.__contents = io.BytesIO(data)
-        self.set_blob_metadata(dict(size=len(data)))
+        metadata = metadata or {}
+        metadata.update({"size": len(data)})
+        self.set_blob_metadata(metadata)
         self._container_client.register_blob_client(self)
+        self._staged_contents = {}
 
     def delete_blob(self):
         self._container_client.delete_blob(self)
@@ -526,6 +529,41 @@ class ReaderTest(unittest.TestCase):
 
         self.assertEqual(data, content)
 
+    def test_read_container_client(self):
+        content = "spirits in the material world".encode("utf-8")
+        blob_name = "test_read_container_client_%s" % BLOB_NAME
+        put_to_container(blob_name, contents=content)
+
+        container_client = CLIENT.get_container_client(CONTAINER_NAME)
+
+        with smart_open.azure.Reader(CONTAINER_NAME, blob_name, container_client) as fin:
+            data = fin.read(100)
+
+        assert data == content
+
+    def test_read_blob_client(self):
+        content = "walking on the moon".encode("utf-8")
+        blob_name = "test_read_blob_client_%s" % BLOB_NAME
+        put_to_container(blob_name, contents=content)
+
+        container_client = CLIENT.get_container_client(CONTAINER_NAME)
+        blob_client = container_client.get_blob_client(blob_name)
+
+        with smart_open.azure.Reader(CONTAINER_NAME, blob_name, blob_client) as fin:
+            data = fin.read(100)
+
+        assert data == content
+
+    def test_nonexisting_container(self):
+        with self.assertRaises(azure.core.exceptions.ResourceNotFoundError):
+            with smart_open.azure.open(
+                    'thiscontainerdoesntexist',
+                    'mykey',
+                    'rb',
+                    CLIENT
+            ) as fin:
+                fin.read()
+
 
 class WriterTest(unittest.TestCase):
     """Test writing into Azure Blob files."""
@@ -548,6 +586,109 @@ class WriterTest(unittest.TestCase):
         ))
         self.assertEqual(output, [test_string])
 
+    def test_write_container_client(self):
+        """Does writing into Azure Blob Storage work correctly?"""
+        test_string = u"Hiszékeny Öngyilkos Vasárnap".encode('utf8')
+        blob_name = "test_write_container_client_%s" % BLOB_NAME
+
+        container_client = CLIENT.get_container_client(CONTAINER_NAME)
+
+        with smart_open.azure.Writer(CONTAINER_NAME, blob_name, container_client) as fout:
+            fout.write(test_string)
+
+        output = list(smart_open.open(
+            "azure://%s/%s" % (CONTAINER_NAME, blob_name),
+            "rb",
+            transport_params=dict(client=container_client),
+        ))
+        assert output == [test_string]
+
+    def test_write_blob_client(self):
+        """Does writing into Azure Blob Storage work correctly?"""
+        test_string = u"žluťoučký koníček".encode('utf8')
+        blob_name = "test_write_blob_client_%s" % BLOB_NAME
+
+        container_client = CLIENT.get_container_client(CONTAINER_NAME)
+        blob_client = container_client.get_blob_client(blob_name)
+
+        with smart_open.open(
+            "azure://%s/%s" % (CONTAINER_NAME, blob_name),
+            "wb",
+            transport_params={
+                "client": blob_client, "blob_kwargs": {"metadata": {"name": blob_name}}
+            },
+        ) as fout:
+            fout.write(test_string)
+
+        self.assertEqual(blob_client.get_blob_properties()["name"], blob_name)
+
+        output = list(smart_open.open(
+            "azure://%s/%s" % (CONTAINER_NAME, blob_name),
+            "rb",
+            transport_params=dict(client=CLIENT),
+        ))
+        self.assertEqual(output, [test_string])
+
+    def test_abort_upload(self):
+        """Does aborted upload skip commit_block_list?"""
+        test_string = b"42" * 42
+        blob_name = "test_abort_upload_%s" % BLOB_NAME
+
+        container_client = CLIENT.get_container_client(CONTAINER_NAME)
+        blob_client = container_client.get_blob_client(blob_name)
+
+        try:
+            with smart_open.open(
+                "azure://%s/%s" % (CONTAINER_NAME, blob_name),
+                "wb",
+                transport_params={"client": blob_client, "min_part_size": 42},
+            ) as fout:
+                fout.write(test_string)
+                raise ValueError
+        except ValueError:
+            # FakeBlobClient.commit_block_list was not called
+            self.assertGreater(len(blob_client._staged_contents), 0)
+
+    def test_abort_upload_text_mode(self):
+        """Does aborted upload skip commit_block_list in text mode?"""
+        test_string = "42" * 42
+        blob_name = "test_abort_upload_%s" % BLOB_NAME
+
+        container_client = CLIENT.get_container_client(CONTAINER_NAME)
+        blob_client = container_client.get_blob_client(blob_name)
+
+        try:
+            with smart_open.open(
+                "azure://%s/%s" % (CONTAINER_NAME, blob_name),
+                "w",
+                transport_params={"client": blob_client, "min_part_size": 42},
+            ) as fout:
+                fout.write(test_string)
+                raise ValueError
+        except ValueError:
+            # FakeBlobClient.commit_block_list was not called
+            self.assertGreater(len(blob_client._staged_contents), 0)
+
+    def test_abort_upload_compressed(self):
+        """Does aborted upload skip commit_block_list with compression?"""
+        test_string = b"42" * 42
+        blob_name = "test_abort_upload_%s.gz" % BLOB_NAME
+
+        container_client = CLIENT.get_container_client(CONTAINER_NAME)
+        blob_client = container_client.get_blob_client(blob_name)
+
+        try:
+            with smart_open.open(
+                "azure://%s/%s" % (CONTAINER_NAME, blob_name),
+                "wb",
+                transport_params={"client": blob_client, "min_part_size": 42},
+            ) as fout:
+                fout.write(test_string)
+                raise ValueError
+        except ValueError:
+            # FakeBlobClient.commit_block_list was not called
+            self.assertGreater(len(blob_client._staged_contents), 0)
+
     def test_incorrect_input(self):
         """Does azure write fail on incorrect input?"""
         blob_name = "test_incorrect_input_%s" % BLOB_NAME
