@@ -6,12 +6,12 @@
 # from the MIT License (MIT).
 #
 from collections import defaultdict
+import functools
 import gzip
 import io
 import logging
 import os
 import tempfile
-import time
 import unittest
 import warnings
 from contextlib import contextmanager
@@ -21,9 +21,16 @@ import sys
 import boto3
 import botocore.client
 import botocore.endpoint
-import moto
+import botocore.exceptions
 import pytest
 
+# See https://github.com/piskvorky/smart_open/issues/800
+# This supports moto 4 & 5 until v4 is no longer used by distros.
+try:
+    from moto import mock_s3
+except ImportError:
+    from moto import mock_aws as mock_s3
+
 import smart_open
 import smart_open.s3
 
@@ -45,42 +52,9 @@ ENABLE_MOTO_SERVER = os.environ.get("SO_ENABLE_MOTO_SERVER") == "1"
 os.environ["AWS_ACCESS_KEY_ID"] = "test"
 os.environ["AWS_SECRET_ACCESS_KEY"] = "test"
 
-
 logger = logging.getLogger(__name__)
 
-
-@moto.mock_s3
-def setUpModule():
-    '''Called once by unittest when initializing this module.  Sets up the
-    test S3 bucket.
-
-    '''
-    bucket = boto3.resource('s3').create_bucket(Bucket=BUCKET_NAME)
-    bucket.wait_until_exists()
-
-
-def cleanup_bucket():
-    for key in boto3.resource('s3').Bucket(BUCKET_NAME).objects.all():
-        key.delete()
-
-
-def put_to_bucket(contents, num_attempts=12, sleep_time=5):
-    logger.debug('%r', locals())
-
-    #
-    # In real life, it can take a few seconds for the bucket to become ready.
-    # If we try to write to the key while the bucket while it isn't ready, we
-    # will get a ClientError: NoSuchBucket.
-    #
-    for attempt in range(num_attempts):
-        try:
-            boto3.resource('s3').Object(BUCKET_NAME, KEY_NAME).put(Body=contents)
-            return
-        except botocore.exceptions.ClientError as err:
-            logger.error('caught %r, retrying', err)
-            time.sleep(sleep_time)
-
-    assert False, 'failed to write to bucket %s after %d attempts' % (BUCKET_NAME, num_attempts)
+_resource = functools.partial(boto3.resource, region_name='us-east-1')
 
 
 def ignore_resource_warnings():
@@ -106,6 +80,8 @@ def patch_invalid_range_response(actual_size):
                 error_response['ActualObjectSize'] = actual_size
                 error_response['Code'] = 'InvalidRange'
                 error_response['Message'] = 'The requested range is not satisfiable'
+            if actual_size is None:
+                error_response.pop('ActualObjectSize', None)
             raise
 
     with mock.patch('smart_open.s3._get', new=mock_get):
@@ -181,7 +157,7 @@ class CrapClient:
             'ContentLength': self._datasize,
             'ContentRange': 'bytes 0-%d/%d' % (self._datasize, self._datasize),
             'Body': self._body,
-            'ResponseMetadata': {'RetryAttempts': 1},
+            'ResponseMetadata': {'RetryAttempts': 1, 'HTTPStatusCode': 206},
         }
 
 
@@ -220,77 +196,68 @@ class IncrementalBackoffTest(unittest.TestCase):
             mock_sleep.reset_mock()
 
 
-@moto.mock_s3
+@mock_s3
 class ReaderTest(BaseTest):
     def setUp(self):
         # lower the multipart upload size, to speed up these tests
-        self.old_min_part_size = smart_open.s3.DEFAULT_MIN_PART_SIZE
-        smart_open.s3.DEFAULT_MIN_PART_SIZE = 5 * 1024**2
+        self.old_min_part_size = smart_open.s3.DEFAULT_PART_SIZE
+        smart_open.s3.DEFAULT_PART_SIZE = 5 * 1024**2
 
         ignore_resource_warnings()
 
         super().setUp()
 
+        s3 = _resource('s3')
+        s3.create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
+
+        self.body = u"hello wořld\nhow are you?".encode('utf8')
+        s3.Object(BUCKET_NAME, KEY_NAME).put(Body=self.body)
+
     def tearDown(self):
-        smart_open.s3.DEFAULT_MIN_PART_SIZE = self.old_min_part_size
-        cleanup_bucket()
+        smart_open.s3.DEFAULT_PART_SIZE = self.old_min_part_size
 
     def test_iter(self):
         """Are S3 files iterated over correctly?"""
-        # a list of strings to test with
-        expected = u"hello wořld\nhow are you?".encode('utf8')
-        put_to_bucket(contents=expected)
-
         # connect to fake s3 and read from the fake key we filled above
         with self.assertApiCalls(GetObject=1):
             fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME)
             output = [line.rstrip(b'\n') for line in fin]
-        self.assertEqual(output, expected.split(b'\n'))
+        self.assertEqual(output, self.body.split(b'\n'))
 
     def test_iter_context_manager(self):
         # same thing but using a context manager
-        expected = u"hello wořld\nhow are you?".encode('utf8')
-        put_to_bucket(contents=expected)
+        _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
+
         with self.assertApiCalls(GetObject=1):
             with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME) as fin:
                 output = [line.rstrip(b'\n') for line in fin]
-        self.assertEqual(output, expected.split(b'\n'))
+        self.assertEqual(output, self.body.split(b'\n'))
 
     def test_read(self):
         """Are S3 files read correctly?"""
-        content = u"hello wořld\nhow are you?".encode('utf8')
-        put_to_bucket(contents=content)
-        logger.debug('content: %r len: %r', content, len(content))
-
         with self.assertApiCalls(GetObject=1):
             fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME)
-            self.assertEqual(content[:6], fin.read(6))
-            self.assertEqual(content[6:14], fin.read(8))  # ř is 2 bytes
-            self.assertEqual(content[14:], fin.read())  # read the rest
+            self.assertEqual(self.body[:6], fin.read(6))
+            self.assertEqual(self.body[6:14], fin.read(8))  # ř is 2 bytes
+            self.assertEqual(self.body[14:], fin.read())  # read the rest
 
     def test_seek_beginning(self):
         """Does seeking to the beginning of S3 files work correctly?"""
-        content = u"hello wořld\nhow are you?".encode('utf8')
-        put_to_bucket(contents=content)
-
         with self.assertApiCalls(GetObject=1):
             fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME)
-            self.assertEqual(content[:6], fin.read(6))
-            self.assertEqual(content[6:14], fin.read(8))  # ř is 2 bytes
+            self.assertEqual(self.body[:6], fin.read(6))
+            self.assertEqual(self.body[6:14], fin.read(8))  # ř is 2 bytes
 
         with self.assertApiCalls(GetObject=1):
             fin.seek(0)
-            self.assertEqual(content, fin.read())  # no size given => read whole file
+            self.assertEqual(self.body, fin.read())  # no size given => read whole file
 
         with self.assertApiCalls(GetObject=1):
             fin.seek(0)
-            self.assertEqual(content, fin.read(-1))  # same thing
+            self.assertEqual(self.body, fin.read(-1))  # same thing
 
     def test_seek_start(self):
         """Does seeking from the start of S3 files work correctly?"""
-        content = u"hello wořld\nhow are you?".encode('utf8')
-        put_to_bucket(contents=content)
-
         with self.assertApiCalls(GetObject=1):
             fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True)
             seek = fin.seek(6)
@@ -300,9 +267,6 @@ class ReaderTest(BaseTest):
 
     def test_seek_current(self):
         """Does seeking from the middle of S3 files work correctly?"""
-        content = u"hello wořld\nhow are you?".encode('utf8')
-        put_to_bucket(contents=content)
-
         with self.assertApiCalls(GetObject=1):
             fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME)
             self.assertEqual(fin.read(5), b'hello')
@@ -314,33 +278,24 @@ class ReaderTest(BaseTest):
 
     def test_seek_end(self):
         """Does seeking from the end of S3 files work correctly?"""
-        content = u"hello wořld\nhow are you?".encode('utf8')
-        put_to_bucket(contents=content)
-
         with self.assertApiCalls(GetObject=1):
             fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True)
             seek = fin.seek(-4, whence=smart_open.constants.WHENCE_END)
-            self.assertEqual(seek, len(content) - 4)
+            self.assertEqual(seek, len(self.body) - 4)
             self.assertEqual(fin.read(), b'you?')
 
     def test_seek_past_end(self):
-        content = u"hello wořld\nhow are you?".encode('utf8')
-        put_to_bucket(contents=content)
-
-        with self.assertApiCalls(GetObject=1), patch_invalid_range_response(str(len(content))):
+        with self.assertApiCalls(GetObject=1), patch_invalid_range_response(str(len(self.body))):
             fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True)
             seek = fin.seek(60)
-            self.assertEqual(seek, len(content))
+            self.assertEqual(seek, len(self.body))
 
     def test_detect_eof(self):
-        content = u"hello wořld\nhow are you?".encode('utf8')
-        put_to_bucket(contents=content)
-
         with self.assertApiCalls(GetObject=1):
             fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME)
             fin.read()
             eof = fin.tell()
-            self.assertEqual(eof, len(content))
+            self.assertEqual(eof, len(self.body))
             fin.seek(0, whence=smart_open.constants.WHENCE_END)
             self.assertEqual(eof, fin.tell())
             fin.seek(eof)
@@ -352,7 +307,8 @@ class ReaderTest(BaseTest):
         buf.close = lambda: None  # keep buffer open so that we can .getvalue()
         with gzip.GzipFile(fileobj=buf, mode='w') as zipfile:
             zipfile.write(expected)
-        put_to_bucket(contents=buf.getvalue())
+
+        _resource('s3').Object(BUCKET_NAME, KEY_NAME).put(Body=buf.getvalue())
 
         #
         # Make sure we're reading things correctly.
@@ -377,7 +333,7 @@ class ReaderTest(BaseTest):
 
     def test_readline(self):
         content = b'englishman\nin\nnew\nyork\n'
-        put_to_bucket(contents=content)
+        _resource('s3').Object(BUCKET_NAME, KEY_NAME).put(Body=content)
 
         with self.assertApiCalls(GetObject=2):
             with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME) as fin:
@@ -393,7 +349,7 @@ class ReaderTest(BaseTest):
 
     def test_readline_tiny_buffer(self):
         content = b'englishman\nin\nnew\nyork\n'
-        put_to_bucket(contents=content)
+        _resource('s3').Object(BUCKET_NAME, KEY_NAME).put(Body=content)
 
         with self.assertApiCalls(GetObject=1):
             with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, buffer_size=8) as fin:
@@ -403,9 +359,6 @@ class ReaderTest(BaseTest):
         self.assertEqual(expected, actual)
 
     def test_read0_does_not_return_data(self):
-        content = b'englishman\nin\nnew\nyork\n'
-        put_to_bucket(contents=content)
-
         with self.assertApiCalls():
             # set defer_seek to verify that read(0) doesn't trigger an unnecessary API call
             with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True) as fin:
@@ -414,28 +367,52 @@ class ReaderTest(BaseTest):
         self.assertEqual(data, b'')
 
     def test_to_boto3(self):
-        contents = b'the spice melange\n'
-        put_to_bucket(contents=contents)
-
         with self.assertApiCalls():
             # set defer_seek to verify that to_boto3() doesn't trigger an unnecessary API call
             with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True) as fin:
-                returned_obj = fin.to_boto3(boto3.resource('s3'))
+                returned_obj = fin.to_boto3(_resource('s3'))
 
         boto3_body = returned_obj.get()['Body'].read()
-        self.assertEqual(contents, boto3_body)
+        self.assertEqual(self.body, boto3_body)
 
     def test_binary_iterator(self):
         expected = u"выйду ночью в поле с конём".encode('utf-8').split(b' ')
-        put_to_bucket(contents=b"\n".join(expected))
+        _resource('s3').Object(BUCKET_NAME, KEY_NAME).put(Body=b'\n'.join(expected))
+
+        # test the __iter__ method
         with self.assertApiCalls(GetObject=1):
             with smart_open.s3.open(BUCKET_NAME, KEY_NAME, 'rb') as fin:
                 actual = [line.rstrip() for line in fin]
         self.assertEqual(expected, actual)
 
+        # test the __next__ method
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.s3.open(BUCKET_NAME, KEY_NAME, 'rb') as fin:
+                first = next(fin).rstrip()
+        self.assertEqual(expected[0], first)
+
+    def test_text_iterator(self):
+        expected = u"выйду ночью в поле с конём".split(' ')
+        uri = f's3://{BUCKET_NAME}/{KEY_NAME}.gz'
+
+        with smart_open.open(uri, 'w', encoding='utf-8') as fout:
+            fout.write('\n'.join(expected))
+
+        # test the __iter__ method
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.open(uri, 'r', encoding='utf-8') as fin:
+                actual = [line.rstrip() for line in fin]
+        self.assertEqual(expected, actual)
+
+        # test the __next__ method
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.open(uri, 'r', encoding='utf-8') as fin:
+                first = next(fin).rstrip()
+        self.assertEqual(expected[0], first)
+
     def test_defer_seek(self):
         content = b'englishman\nin\nnew\nyork\n'
-        put_to_bucket(contents=content)
+        _resource('s3').Object(BUCKET_NAME, KEY_NAME).put(Body=content)
 
         with self.assertApiCalls():
             fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True)
@@ -449,7 +426,7 @@ class ReaderTest(BaseTest):
             self.assertEqual(fin.read(), content[10:])
 
     def test_read_empty_file(self):
-        put_to_bucket(contents=b'')
+        _resource('s3').Object(BUCKET_NAME, KEY_NAME).put(Body=b'')
 
         with self.assertApiCalls(GetObject=1), patch_invalid_range_response('0'):
             with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME) as fin:
@@ -457,8 +434,17 @@ class ReaderTest(BaseTest):
 
         self.assertEqual(data, b'')
 
+    def test_read_empty_file_no_actual_size(self):
+        _resource('s3').Object(BUCKET_NAME, KEY_NAME).put(Body=b'')
 
-@moto.mock_s3
+        with self.assertApiCalls(GetObject=2), patch_invalid_range_response(None):
+            with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME) as fin:
+                data = fin.read()
+
+        self.assertEqual(data, b'')
+
+
+@mock_s3
 class MultipartWriterTest(unittest.TestCase):
     """
     Test writing into s3 files.
@@ -467,8 +453,7 @@ class MultipartWriterTest(unittest.TestCase):
     def setUp(self):
         ignore_resource_warnings()
 
-    def tearDown(self):
-        cleanup_bucket()
+        _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
 
     def test_write_01(self):
         """Does writing into s3 work correctly?"""
@@ -502,27 +487,63 @@ class MultipartWriterTest(unittest.TestCase):
             fout.write(u"testžížáč".encode("utf-8"))
             self.assertEqual(fout.tell(), 14)
 
+    #
+    # Nb. Under Windows, the byte offsets are different for some reason
+    #
+    @pytest.mark.skipif(condition=sys.platform == 'win32', reason="does not run on windows")
     def test_write_03(self):
         """Does s3 multipart chunking work correctly?"""
-        # write
-        smart_open_write = smart_open.s3.MultipartWriter(
-            BUCKET_NAME, WRITE_KEY_NAME, min_part_size=10
-        )
-        with smart_open_write as fout:
-            fout.write(b"test")
-            self.assertEqual(fout._buf.tell(), 4)
+        #
+        # generate enough test data for a single multipart upload part.
+        # We need this because moto behaves like S3; it refuses to upload
+        # parts smaller than 5MB.
+        #
+        data_dir = os.path.join(os.path.dirname(__file__), "test_data")
+        with open(os.path.join(data_dir, "crime-and-punishment.txt"), "rb") as fin:
+            crime = fin.read()
+        data = b''
+        ps = 5 * 1024 * 1024
+        while len(data) < ps:
+            data += crime
 
-            fout.write(b"test\n")
-            self.assertEqual(fout._buf.tell(), 9)
-            self.assertEqual(fout._total_parts, 0)
+        title = "Преступление и наказание\n\n".encode()
+        to_be_continued = "\n\n... продолжение следует ...\n\n".encode()
 
-            fout.write(b"test")
-            self.assertEqual(fout._buf.tell(), 0)
-            self.assertEqual(fout._total_parts, 1)
+        with smart_open.s3.MultipartWriter(BUCKET_NAME, WRITE_KEY_NAME, part_size=ps) as fout:
+            #
+            # Write some data without triggering an upload
+            #
+            fout.write(title)
+            assert fout._total_parts == 0
+            assert fout._buf.tell() == 48
 
+            #
+            # Trigger a part upload
+            #
+            fout.write(data)
+            assert fout._total_parts == 1
+            assert fout._buf.tell() == 661
+
+            #
+            # Write _without_ triggering a part upload
+            #
+            fout.write(to_be_continued)
+            assert fout._total_parts == 1
+            assert fout._buf.tell() == 710
+
+        #
+        # We closed the writer, so the final part must have been uploaded
+        #
+        assert fout._buf.tell() == 0
+        assert fout._total_parts == 2
+
+        #
         # read back the same key and check its content
-        output = list(smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb'))
-        self.assertEqual(output, [b"testtest\n", b"test"])
+        #
+        with smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb') as fin:
+            got = fin.read()
+        want = title + data + to_be_continued
+        assert want == got
 
     def test_write_04(self):
         """Does writing no data cause key with an empty value to be created?"""
@@ -595,7 +616,7 @@ class MultipartWriterTest(unittest.TestCase):
 
         with smart_open.s3.open(BUCKET_NAME, KEY_NAME, 'wb') as fout:
             fout.write(contents)
-            returned_obj = fout.to_boto3(boto3.resource('s3'))
+            returned_obj = fout.to_boto3(_resource('s3'))
 
         boto3_body = returned_obj.get()['Body'].read()
         self.assertEqual(contents, boto3_body)
@@ -613,8 +634,95 @@ class MultipartWriterTest(unittest.TestCase):
 
             assert actual == contents
 
+    def test_write_gz_using_context_manager(self):
+        """Does s3 multipart upload create a compressed file using context manager?"""
+        contents = b'get ready for a surprise'
+        with smart_open.open(
+                f's3://{BUCKET_NAME}/{WRITE_KEY_NAME}.gz',
+                mode="wb",
+                transport_params={
+                    "multipart_upload": True,
+                    "min_part_size": 10,
+                }
+        ) as fout:
+            fout.write(contents)
 
-@moto.mock_s3
+        with smart_open.open(f's3://{BUCKET_NAME}/{WRITE_KEY_NAME}.gz', 'rb') as fin:
+            actual = fin.read()
+
+        assert actual == contents
+
+    def test_write_gz_not_using_context_manager(self):
+        """Does s3 multipart upload create a compressed file not using context manager but close()?"""
+        contents = b'get ready for a surprise'
+        fout = smart_open.open(
+            f's3://{BUCKET_NAME}/{WRITE_KEY_NAME}.gz',
+            mode="wb",
+            transport_params={
+                "multipart_upload": True,
+                "min_part_size": 10,
+            }
+        )
+        fout.write(contents)
+        fout.close()
+
+        with smart_open.open(f's3://{BUCKET_NAME}/{WRITE_KEY_NAME}.gz', 'rb') as fin:
+            actual = fin.read()
+
+        assert actual == contents
+
+    def test_write_gz_with_error(self):
+        """Does s3 multipart upload abort for a failed compressed file upload?"""
+        with self.assertRaises(ValueError):
+            with smart_open.open(
+                    f's3://{BUCKET_NAME}/{WRITE_KEY_NAME}',
+                    mode="wb",
+                    compression='.gz',
+                    transport_params={
+                        "multipart_upload": True,
+                        "min_part_size": 10,
+                    }
+            ) as fout:
+                fout.write(b"test12345678test12345678")
+                fout.write(b"test\n")
+
+                # FileLikeWrapper.__exit__ should cause a MultipartWriter.terminate()
+                raise ValueError("some error")
+
+        # no multipart upload was committed:
+        # smart_open.s3.MultipartWriter.__exit__ was called
+        with self.assertRaises(OSError) as cm:
+            smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb')
+
+        assert 'The specified key does not exist.' in cm.exception.args[0]
+
+    def test_write_text_with_error(self):
+        """Does s3 multipart upload abort for a failed text file upload?"""
+        with self.assertRaises(ValueError):
+            with smart_open.open(
+                    f's3://{BUCKET_NAME}/{WRITE_KEY_NAME}',
+                    mode="w",
+                    transport_params={
+                        "multipart_upload": True,
+                        "min_part_size": 10,
+                    }
+            ) as fout:
+                fout.write("test12345678test12345678")
+                fout.write("test\n")
+
+                # TextIOWrapper.__exit__ should not cause a self.buffer.close()
+                # FileLikeWrapper.__exit__ should cause a MultipartWriter.terminate()
+                raise ValueError("some error")
+
+        # no multipart upload was committed:
+        # smart_open.s3.MultipartWriter.__exit__ was called
+        with self.assertRaises(OSError) as cm:
+            smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb')
+
+        assert 'The specified key does not exist.' in cm.exception.args[0]
+
+
+@mock_s3
 class SinglepartWriterTest(unittest.TestCase):
     """
     Test writing into s3 files using single part upload.
@@ -623,8 +731,7 @@ class SinglepartWriterTest(unittest.TestCase):
     def setUp(self):
         ignore_resource_warnings()
 
-    def tearDown(self):
-        cleanup_bucket()
+        _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
 
     def test_write_01(self):
         """Does writing into s3 work correctly?"""
@@ -721,17 +828,39 @@ class SinglepartWriterTest(unittest.TestCase):
 
             assert actual == contents
 
+    def test_str(self):
+        with smart_open.s3.open(BUCKET_NAME, 'key', 'wb', multipart_upload=False) as fout:
+            assert str(fout) == "smart_open.s3.SinglepartWriter('test-smartopen', 'key')"
+
+    def test_ensure_no_side_effects_on_exception(self):
+        class WriteError(Exception):
+            pass
+
+        s3_resource = _resource("s3")
+        obj = s3_resource.Object(BUCKET_NAME, KEY_NAME)
+
+        # wrap in closure to ease writer dereferencing
+        def _run():
+            with smart_open.s3.open(BUCKET_NAME, obj.key, "wb", multipart_upload=False) as fout:
+                fout.write(b"this should not be written")
+                raise WriteError
+
+        try:
+            _run()
+        except WriteError:
+            pass
+        finally:
+            self.assertRaises(s3_resource.meta.client.exceptions.NoSuchKey, obj.get)
+
 
 ARBITRARY_CLIENT_ERROR = botocore.client.ClientError(error_response={}, operation_name='bar')
 
 
-@moto.mock_s3
+@mock_s3
 class IterBucketTest(unittest.TestCase):
     def setUp(self):
         ignore_resource_warnings()
-
-    def tearDown(self):
-        cleanup_bucket()
+        _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
 
     @pytest.mark.skipif(condition=sys.platform == 'win32', reason="does not run on windows")
     @pytest.mark.xfail(
@@ -743,6 +872,38 @@ class IterBucketTest(unittest.TestCase):
         results = list(smart_open.s3.iter_bucket(BUCKET_NAME))
         self.assertEqual(len(results), 10)
 
+    @pytest.mark.skipif(condition=sys.platform == 'win32', reason="does not run on windows")
+    @pytest.mark.xfail(
+        condition=sys.platform == 'darwin',
+        reason="MacOS uses spawn rather than fork for multiprocessing",
+    )
+    def test_iter_bucket_404(self):
+        populate_bucket()
+
+        def throw_404_error_for_key_4(*args):
+            if args[1] == "key_4":
+                raise botocore.exceptions.ClientError(
+                    error_response={"Error": {"Code": "404", "Message": "Not Found"}},
+                    operation_name="HeadObject",
+                )
+            else:
+                return [0]
+
+        with mock.patch("smart_open.s3._download_fileobj", side_effect=throw_404_error_for_key_4):
+            results = list(smart_open.s3.iter_bucket(BUCKET_NAME))
+            self.assertEqual(len(results), 9)
+
+    @pytest.mark.skipif(condition=sys.platform == 'win32', reason="does not run on windows")
+    @pytest.mark.xfail(
+        condition=sys.platform == 'darwin',
+        reason="MacOS uses spawn rather than fork for multiprocessing",
+    )
+    def test_iter_bucket_non_404(self):
+        populate_bucket()
+        with mock.patch("smart_open.s3._download_fileobj", side_effect=ARBITRARY_CLIENT_ERROR):
+            with pytest.raises(botocore.exceptions.ClientError):
+                list(smart_open.s3.iter_bucket(BUCKET_NAME))
+
     def test_deprecated_top_level_s3_iter_bucket(self):
         populate_bucket()
         with self.assertLogs(smart_open.logger.name, level='WARN') as cm:
@@ -762,7 +923,7 @@ class IterBucketTest(unittest.TestCase):
     )
     def test_accepts_boto3_bucket(self):
         populate_bucket()
-        bucket = boto3.resource('s3').Bucket(BUCKET_NAME)
+        bucket = _resource('s3').Bucket(BUCKET_NAME)
         results = list(smart_open.s3.iter_bucket(bucket))
         self.assertEqual(len(results), 10)
 
@@ -785,7 +946,7 @@ class IterBucketTest(unittest.TestCase):
         self.assertEqual(sorted(keys), sorted(expected))
 
 
-@moto.mock_s3
+@mock_s3
 @pytest.mark.skipif(
     condition=not smart_open.concurrency._CONCURRENT_FUTURES,
     reason='concurrent.futures unavailable',
@@ -801,9 +962,10 @@ class IterBucketConcurrentFuturesTest(unittest.TestCase):
         smart_open.concurrency._MULTIPROCESSING = False
         ignore_resource_warnings()
 
+        _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
+
     def tearDown(self):
         smart_open.concurrency._MULTIPROCESSING = self.old_flag_multi
-        cleanup_bucket()
 
     def test(self):
         num_keys = 101
@@ -815,7 +977,7 @@ class IterBucketConcurrentFuturesTest(unittest.TestCase):
         self.assertEqual(sorted(keys), sorted(expected))
 
 
-@moto.mock_s3
+@mock_s3
 @pytest.mark.skipif(
     condition=not smart_open.concurrency._MULTIPROCESSING,
     reason='multiprocessing unavailable',
@@ -831,9 +993,10 @@ class IterBucketMultiprocessingTest(unittest.TestCase):
         smart_open.concurrency._CONCURRENT_FUTURES = False
         ignore_resource_warnings()
 
+        _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
+
     def tearDown(self):
         smart_open.concurrency._CONCURRENT_FUTURES = self.old_flag_concurrent
-        cleanup_bucket()
 
     def test(self):
         num_keys = 101
@@ -845,7 +1008,7 @@ class IterBucketMultiprocessingTest(unittest.TestCase):
         self.assertEqual(sorted(keys), sorted(expected))
 
 
-@moto.mock_s3
+@mock_s3
 class IterBucketSingleProcessTest(unittest.TestCase):
     def setUp(self):
         self.old_flag_multi = smart_open.concurrency._MULTIPROCESSING
@@ -855,10 +1018,11 @@ class IterBucketSingleProcessTest(unittest.TestCase):
 
         ignore_resource_warnings()
 
+        _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
+
     def tearDown(self):
         smart_open.concurrency._MULTIPROCESSING = self.old_flag_multi
         smart_open.concurrency._CONCURRENT_FUTURES = self.old_flag_concurrent
-        cleanup_bucket()
 
     def test(self):
         num_keys = 101
@@ -874,9 +1038,10 @@ class IterBucketSingleProcessTest(unittest.TestCase):
 # This has to be a separate test because we cannot run it against real S3
 # (we don't want to expose our real S3 credentials).
 #
-@moto.mock_s3
+@mock_s3
 class IterBucketCredentialsTest(unittest.TestCase):
     def test(self):
+        _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
         num_keys = 10
         populate_bucket(num_keys=num_keys)
         result = list(
@@ -890,33 +1055,31 @@ class IterBucketCredentialsTest(unittest.TestCase):
         self.assertEqual(len(result), num_keys)
 
 
-@moto.mock_s3
+@mock_s3
 class DownloadKeyTest(unittest.TestCase):
     def setUp(self):
         ignore_resource_warnings()
 
-    def tearDown(self):
-        cleanup_bucket()
+        s3 = _resource('s3')
+        bucket = s3.create_bucket(Bucket=BUCKET_NAME)
+        bucket.wait_until_exists()
+
+        self.body = b'hello'
+        s3.Object(BUCKET_NAME, KEY_NAME).put(Body=self.body)
 
     def test_happy(self):
-        contents = b'hello'
-        put_to_bucket(contents=contents)
-        expected = (KEY_NAME, contents)
+        expected = (KEY_NAME, self.body)
         actual = smart_open.s3._download_key(KEY_NAME, bucket_name=BUCKET_NAME)
         self.assertEqual(expected, actual)
 
     def test_intermittent_error(self):
-        contents = b'hello'
-        put_to_bucket(contents=contents)
-        expected = (KEY_NAME, contents)
-        side_effect = [ARBITRARY_CLIENT_ERROR, ARBITRARY_CLIENT_ERROR, contents]
+        expected = (KEY_NAME, self.body)
+        side_effect = [ARBITRARY_CLIENT_ERROR, ARBITRARY_CLIENT_ERROR, self.body]
         with mock.patch('smart_open.s3._download_fileobj', side_effect=side_effect):
             actual = smart_open.s3._download_key(KEY_NAME, bucket_name=BUCKET_NAME)
         self.assertEqual(expected, actual)
 
     def test_persistent_error(self):
-        contents = b'hello'
-        put_to_bucket(contents=contents)
         side_effect = [ARBITRARY_CLIENT_ERROR, ARBITRARY_CLIENT_ERROR,
                        ARBITRARY_CLIENT_ERROR, ARBITRARY_CLIENT_ERROR]
         with mock.patch('smart_open.s3._download_fileobj', side_effect=side_effect):
@@ -924,30 +1087,24 @@ class DownloadKeyTest(unittest.TestCase):
                               KEY_NAME, bucket_name=BUCKET_NAME)
 
     def test_intermittent_error_retries(self):
-        contents = b'hello'
-        put_to_bucket(contents=contents)
-        expected = (KEY_NAME, contents)
+        expected = (KEY_NAME, self.body)
         side_effect = [ARBITRARY_CLIENT_ERROR, ARBITRARY_CLIENT_ERROR,
-                       ARBITRARY_CLIENT_ERROR, ARBITRARY_CLIENT_ERROR, contents]
+                       ARBITRARY_CLIENT_ERROR, ARBITRARY_CLIENT_ERROR, self.body]
         with mock.patch('smart_open.s3._download_fileobj', side_effect=side_effect):
             actual = smart_open.s3._download_key(KEY_NAME, bucket_name=BUCKET_NAME, retries=4)
         self.assertEqual(expected, actual)
 
     def test_propagates_other_exception(self):
-        contents = b'hello'
-        put_to_bucket(contents=contents)
         with mock.patch('smart_open.s3._download_fileobj', side_effect=ValueError):
             self.assertRaises(ValueError, smart_open.s3._download_key,
                               KEY_NAME, bucket_name=BUCKET_NAME)
 
 
-@moto.mock_s3
+@mock_s3
 class OpenTest(unittest.TestCase):
     def setUp(self):
         ignore_resource_warnings()
-
-    def tearDown(self):
-        cleanup_bucket()
+        _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
 
     def test_read_never_returns_none(self):
         """read should never return None."""
@@ -962,30 +1119,43 @@ class OpenTest(unittest.TestCase):
 
 
 def populate_bucket(num_keys=10):
-    s3 = boto3.resource('s3')
+    s3 = _resource('s3')
     for key_number in range(num_keys):
         key_name = 'key_%d' % key_number
         s3.Object(BUCKET_NAME, key_name).put(Body=str(key_number))
 
 
 class RetryIfFailedTest(unittest.TestCase):
+    def setUp(self):
+        self.retry = smart_open.s3.Retry()
+        self.retry.attempts = 3
+        self.retry.sleep_seconds = 0
+
     def test_success(self):
         partial = mock.Mock(return_value=1)
-        result = smart_open.s3._retry_if_failed(partial, attempts=3, sleep_seconds=0)
+        result = self.retry._do(partial)
         self.assertEqual(result, 1)
         self.assertEqual(partial.call_count, 1)
 
-    def test_failure(self):
+    def test_failure_exception(self):
         partial = mock.Mock(side_effect=ValueError)
-        exceptions = (ValueError, )
-
+        self.retry.exceptions = {ValueError: 'Let us retry ValueError'}
         with self.assertRaises(IOError):
-            smart_open.s3._retry_if_failed(partial, attempts=3, sleep_seconds=0, exceptions=exceptions)
+            self.retry._do(partial)
+        self.assertEqual(partial.call_count, 3)
 
+    def test_failure_client_error(self):
+        partial = mock.Mock(
+            side_effect=botocore.exceptions.ClientError(
+                {'Error': {'Code': 'NoSuchUpload'}}, 'NoSuchUpload'
+            )
+        )
+        with self.assertRaises(IOError):
+            self.retry._do(partial)
         self.assertEqual(partial.call_count, 3)
 
 
-@moto.mock_s3()
+@mock_s3
 def test_client_propagation_singlepart():
     """Does the client parameter make it from the caller to Boto3?"""
     #
@@ -993,9 +1163,7 @@ def test_client_propagation_singlepart():
     # have done that for us by now.
     #
     session = boto3.Session()
-    resource = session.resource('s3')
-    bucket = resource.create_bucket(Bucket=BUCKET_NAME)
-    bucket.wait_until_exists()
+    _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
 
     client = session.client('s3')
 
@@ -1010,13 +1178,11 @@ def test_client_propagation_singlepart():
         assert id(writer._client.client) == id(client)
 
 
-@moto.mock_s3()
+@mock_s3
 def test_client_propagation_multipart():
     """Does the resource parameter make it from the caller to Boto3?"""
     session = boto3.Session()
-    resource = session.resource('s3')
-    bucket = resource.create_bucket(Bucket=BUCKET_NAME)
-    bucket.wait_until_exists()
+    _resource('s3').create_bucket(Bucket=BUCKET_NAME).wait_until_exists()
 
     client = session.client('s3')
 
@@ -1031,11 +1197,11 @@ def test_client_propagation_multipart():
         assert id(writer._client.client) == id(client)
 
 
-@moto.mock_s3()
+@mock_s3
 def test_resource_propagation_reader():
     """Does the resource parameter make it from the caller to Boto3?"""
     session = boto3.Session()
-    resource = session.resource('s3')
+    resource = session.resource('s3', region_name='us-east-1')
     bucket = resource.create_bucket(Bucket=BUCKET_NAME)
     bucket.wait_until_exists()
 
